@@ -1,4 +1,29 @@
-#include "fitz-imp.h"
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
+#include "mupdf/fitz.h"
+
+#include <string.h>
+#include <stdarg.h>
 
 fz_buffer *
 fz_new_buffer(fz_context *ctx, size_t size)
@@ -11,7 +36,7 @@ fz_new_buffer(fz_context *ctx, size_t size)
 	b->refs = 1;
 	fz_try(ctx)
 	{
-		b->data = fz_malloc(ctx, size);
+		b->data = Memento_label(fz_malloc(ctx, size), "fz_buffer_data");
 	}
 	fz_catch(ctx)
 	{
@@ -28,7 +53,7 @@ fz_new_buffer(fz_context *ctx, size_t size)
 fz_buffer *
 fz_new_buffer_from_data(fz_context *ctx, unsigned char *data, size_t size)
 {
-	fz_buffer *b;
+	fz_buffer *b = NULL;
 
 	fz_try(ctx)
 	{
@@ -49,7 +74,7 @@ fz_new_buffer_from_data(fz_context *ctx, unsigned char *data, size_t size)
 }
 
 fz_buffer *
-fz_new_buffer_from_shared_data(fz_context *ctx, const char *data, size_t size)
+fz_new_buffer_from_shared_data(fz_context *ctx, const unsigned char *data, size_t size)
 {
 	fz_buffer *b;
 
@@ -64,35 +89,108 @@ fz_new_buffer_from_shared_data(fz_context *ctx, const char *data, size_t size)
 	return b;
 }
 
+fz_buffer *
+fz_new_buffer_from_copied_data(fz_context *ctx, const unsigned char *data, size_t size)
+{
+	fz_buffer *b;
+	if (size > 0 && data == NULL)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no data provided");
+	b = fz_new_buffer(ctx, size);
+	b->len = size;
+	memcpy(b->data, data, size);
+	return b;
+}
+
+fz_buffer *fz_clone_buffer(fz_context *ctx, fz_buffer *buf)
+{
+	return fz_new_buffer_from_copied_data(ctx, buf ? buf->data : NULL, buf ? buf->len : 0);
+}
+
+static inline int iswhite(int a)
+{
+	switch (a) {
+	case '\n': case '\r': case '\t': case ' ':
+	case '\f':
+		return 1;
+	}
+	return 0;
+}
+
 fz_buffer *
 fz_new_buffer_from_base64(fz_context *ctx, const char *data, size_t size)
 {
-	fz_buffer *buf = fz_new_buffer(ctx, size);
-	const char *end = data + size;
+	fz_buffer *out = fz_new_buffer(ctx, size > 0 ? size : strlen(data));
+	const char *end = data + (size > 0 ? size : strlen(data));
 	const char *s = data;
+	uint32_t buf = 0;
+	int bits = 0;
+
+	/* This is https://infra.spec.whatwg.org/#forgiving-base64-decode
+	 * but even more relaxed. We allow any number of trailing '=' code
+	 * points and instead of returning failure on invalid characters, we
+	 * warn and truncate.
+	 */
+
+	while (s < end && iswhite(*s))
+		s++;
+	while (s < end && iswhite(end[-1]))
+		end--;
+	while (s < end && end[-1] == '=')
+		end--;
+
 	fz_try(ctx)
 	{
 		while (s < end)
 		{
 			int c = *s++;
+
 			if (c >= 'A' && c <= 'Z')
-				fz_append_bits(ctx, buf, c - 'A', 6);
+				c = c - 'A';
 			else if (c >= 'a' && c <= 'z')
-				fz_append_bits(ctx, buf, c - 'a' + 26, 6);
+				c = c - 'a' + 26;
 			else if (c >= '0' && c <= '9')
-				fz_append_bits(ctx, buf, c - '0' + 52, 6);
+				c = c - '0' + 52;
 			else if (c == '+')
-				fz_append_bits(ctx, buf, 62, 6);
+				c = 62;
 			else if (c == '/')
-				fz_append_bits(ctx, buf, 63, 6);
+				c = 63;
+			else if (iswhite(c))
+				continue;
+			else
+			{
+				fz_warn(ctx, "invalid character in base64");
+				break;
+			}
+
+			buf <<= 6;
+			buf |= c & 0x3f;
+			bits += 6;
+
+			if (bits == 24)
+			{
+				fz_append_byte(ctx, out, buf >> 16);
+				fz_append_byte(ctx, out, buf >> 8);
+				fz_append_byte(ctx, out, buf >> 0);
+				bits = 0;
+			}
+		}
+
+		if (bits == 18)
+		{
+			fz_append_byte(ctx, out, buf >> 10);
+			fz_append_byte(ctx, out, buf >> 2);
+		}
+		else if (bits == 12)
+		{
+			fz_append_byte(ctx, out, buf >> 4);
 		}
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_buffer(ctx, buf);
+		fz_drop_buffer(ctx, out);
 		fz_rethrow(ctx);
 	}
-	return buf;
+	return out;
 }
 
 fz_buffer *
@@ -116,8 +214,8 @@ void
 fz_resize_buffer(fz_context *ctx, fz_buffer *buf, size_t size)
 {
 	if (buf->shared)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot resize a buffer with shared storage");
-	buf->data = fz_resize_array(ctx, buf->data, size, 1);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot resize a buffer with shared storage");
+	buf->data = fz_realloc(ctx, buf->data, size);
 	buf->cap = size;
 	if (buf->len > buf->cap)
 		buf->len = buf->cap;
@@ -152,6 +250,12 @@ fz_trim_buffer(fz_context *ctx, fz_buffer *buf)
 		fz_resize_buffer(ctx, buf, buf->len);
 }
 
+void
+fz_clear_buffer(fz_context *ctx, fz_buffer *buf)
+{
+	buf->len = 0;
+}
+
 void
 fz_terminate_buffer(fz_context *ctx, fz_buffer *buf)
 {
@@ -192,12 +296,33 @@ fz_buffer_extract(fz_context *ctx, fz_buffer *buf, unsigned char **datap)
 	return len;
 }
 
+fz_buffer *
+fz_slice_buffer(fz_context *ctx, fz_buffer *buf, int64_t start, int64_t end)
+{
+	unsigned char *src = NULL;
+	size_t size = fz_buffer_storage(ctx, buf, &src);
+	size_t s, e;
+
+	if (start < 0)
+		start += size;
+	if (end < 0)
+		end += size;
+
+	s = fz_clamp64(start, 0, size);
+	e = fz_clamp64(end, 0, size);
+
+	if (s == size || e <= s)
+		return fz_new_buffer(ctx, 0);
+
+	return fz_new_buffer_from_copied_data(ctx, &src[s], e - s);
+}
+
 void
 fz_append_buffer(fz_context *ctx, fz_buffer *buf, fz_buffer *extra)
 {
 	if (buf->cap - buf->len < extra->len)
 	{
-		buf->data = fz_resize_array(ctx, buf->data, buf->len + extra->len, 1);
+		buf->data = fz_realloc(ctx, buf->data, buf->len + extra->len);
 		buf->cap = buf->len + extra->len;
 	}
 
@@ -247,6 +372,22 @@ fz_append_rune(fz_context *ctx, fz_buffer *buf, int c)
 	buf->unused_bits = 0;
 }
 
+void
+fz_append_int32_be(fz_context *ctx, fz_buffer *buf, int x)
+{
+	fz_append_byte(ctx, buf, (x >> 24) & 0xFF);
+	fz_append_byte(ctx, buf, (x >> 16) & 0xFF);
+	fz_append_byte(ctx, buf, (x >> 8) & 0xFF);
+	fz_append_byte(ctx, buf, (x) & 0xFF);
+}
+
+void
+fz_append_int16_be(fz_context *ctx, fz_buffer *buf, int x)
+{
+	fz_append_byte(ctx, buf, (x >> 8) & 0xFF);
+	fz_append_byte(ctx, buf, (x) & 0xFF);
+}
+
 void
 fz_append_int32_le(fz_context *ctx, fz_buffer *buf, int x)
 {
@@ -432,7 +573,8 @@ fz_md5_buffer(fz_context *ctx, fz_buffer *buffer, unsigned char digest[16])
 {
 	fz_md5 state;
 	fz_md5_init(&state);
-	fz_md5_update(&state, buffer->data, buffer->len);
+	if (buffer)
+		fz_md5_update(&state, buffer->data, buffer->len);
 	fz_md5_final(&state, digest);
 }
 
