@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -38,6 +38,8 @@ pdf_obj_num_is_stream(fz_context *ctx, pdf_document *doc, int num)
 	fz_catch(ctx)
 	{
 		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+		fz_report_error(ctx);
 		return 0;
 	}
 
@@ -94,13 +96,14 @@ pdf_load_jbig2_globals(fz_context *ctx, pdf_obj *dict)
 		return globals;
 
 	if (pdf_mark_obj(ctx, dict))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cyclic reference when loading JBIG2 globals");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cyclic reference when loading JBIG2 globals");
 
 	fz_try(ctx)
 	{
 		buf = pdf_load_stream(ctx, dict);
 		globals = fz_load_jbig2_globals(ctx, buf);
-		pdf_store_item(ctx, dict, globals, fz_buffer_storage(ctx, buf, NULL));
+		if (globals)
+			pdf_store_item(ctx, dict, globals, fz_buffer_storage(ctx, buf, NULL));
 	}
 	fz_always(ctx)
 	{
@@ -118,40 +121,24 @@ pdf_load_jbig2_globals(fz_context *ctx, pdf_obj *dict)
 static void
 build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression_params *params)
 {
-	int predictor = pdf_dict_get_int(ctx, p, PDF_NAME(Predictor));
-	pdf_obj *columns_obj = pdf_dict_get(ctx, p, PDF_NAME(Columns));
-	int columns = pdf_to_int(ctx, columns_obj);
-	int colors = pdf_dict_get_int(ctx, p, PDF_NAME(Colors));
-	int bpc = pdf_dict_get_int(ctx, p, PDF_NAME(BitsPerComponent));
-	if (bpc == 0)
-		bpc = 8;
-
 	params->type = FZ_IMAGE_RAW;
 
 	if (pdf_name_eq(ctx, f, PDF_NAME(CCITTFaxDecode)) || pdf_name_eq(ctx, f, PDF_NAME(CCF)))
 	{
-		pdf_obj *k = pdf_dict_get(ctx, p, PDF_NAME(K));
-		pdf_obj *eol = pdf_dict_get(ctx, p, PDF_NAME(EndOfLine));
-		pdf_obj *eba = pdf_dict_get(ctx, p, PDF_NAME(EncodedByteAlign));
-		pdf_obj *rows = pdf_dict_get(ctx, p, PDF_NAME(Rows));
-		pdf_obj *eob = pdf_dict_get(ctx, p, PDF_NAME(EndOfBlock));
-		pdf_obj *bi1 = pdf_dict_get(ctx, p, PDF_NAME(BlackIs1));
-
 		params->type = FZ_IMAGE_FAX;
-		params->u.fax.k = (k ? pdf_to_int(ctx, k) : 0);
-		params->u.fax.end_of_line = (eol ? pdf_to_bool(ctx, eol) : 0);
-		params->u.fax.encoded_byte_align = (eba ? pdf_to_bool(ctx, eba) : 0);
-		params->u.fax.columns = (columns_obj ? columns : 1728);
-		params->u.fax.rows = (rows ? pdf_to_int(ctx, rows) : 0);
-		params->u.fax.end_of_block = (eob ? pdf_to_bool(ctx, eob) : 1);
-		params->u.fax.black_is_1 = (bi1 ? pdf_to_bool(ctx, bi1) : 0);
+		params->u.fax.k = pdf_dict_get_int_default(ctx, p, PDF_NAME(K), 0);
+		params->u.fax.end_of_line = pdf_dict_get_bool_default(ctx, p, PDF_NAME(EndOfLine), 0);
+		params->u.fax.encoded_byte_align = pdf_dict_get_bool_default(ctx, p, PDF_NAME(EncodedByteAlign), 0);
+		params->u.fax.columns = pdf_dict_get_int_default(ctx, p, PDF_NAME(Columns), 1728);
+		params->u.fax.rows = pdf_dict_get_int_default(ctx, p, PDF_NAME(Rows), 0);
+		params->u.fax.end_of_block = pdf_dict_get_bool_default(ctx, p, PDF_NAME(EndOfBlock), 1);
+		params->u.fax.black_is_1 = pdf_dict_get_bool_default(ctx, p, PDF_NAME(BlackIs1), 0);
 	}
 	else if (pdf_name_eq(ctx, f, PDF_NAME(DCTDecode)) || pdf_name_eq(ctx, f, PDF_NAME(DCT)))
 	{
-		pdf_obj *ct = pdf_dict_get(ctx, p, PDF_NAME(ColorTransform));
-
 		params->type = FZ_IMAGE_JPEG;
-		params->u.jpeg.color_transform = (ct ? pdf_to_int(ctx, ct) : -2);
+		params->u.jpeg.color_transform = pdf_dict_get_int_default(ctx, p, PDF_NAME(ColorTransform), -1);
+		params->u.jpeg.invert_cmyk = 0;
 	}
 	else if (pdf_name_eq(ctx, f, PDF_NAME(RunLengthDecode)) || pdf_name_eq(ctx, f, PDF_NAME(RL)))
 	{
@@ -160,21 +147,19 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
 	else if (pdf_name_eq(ctx, f, PDF_NAME(FlateDecode)) || pdf_name_eq(ctx, f, PDF_NAME(Fl)))
 	{
 		params->type = FZ_IMAGE_FLATE;
-		params->u.flate.predictor = predictor;
-		params->u.flate.columns = columns;
-		params->u.flate.colors = colors;
-		params->u.flate.bpc = bpc;
+		params->u.flate.predictor = pdf_dict_get_int_default(ctx, p, PDF_NAME(Predictor), 1);
+		params->u.flate.columns = pdf_dict_get_int_default(ctx, p, PDF_NAME(Columns), 1);
+		params->u.flate.colors = pdf_dict_get_int_default(ctx, p, PDF_NAME(Colors), 1);
+		params->u.flate.bpc = pdf_dict_get_int_default(ctx, p, PDF_NAME(BitsPerComponent), 8);
 	}
 	else if (pdf_name_eq(ctx, f, PDF_NAME(LZWDecode)) || pdf_name_eq(ctx, f, PDF_NAME(LZW)))
 	{
-		pdf_obj *ec = pdf_dict_get(ctx, p, PDF_NAME(EarlyChange));
-
 		params->type = FZ_IMAGE_LZW;
-		params->u.lzw.predictor = predictor;
-		params->u.lzw.columns = columns;
-		params->u.lzw.colors = colors;
-		params->u.lzw.bpc = bpc;
-		params->u.lzw.early_change = (ec ? pdf_to_int(ctx, ec) : 1);
+		params->u.lzw.predictor = pdf_dict_get_int_default(ctx, p, PDF_NAME(Predictor), 1);
+		params->u.lzw.columns = pdf_dict_get_int_default(ctx, p, PDF_NAME(Columns), 1);
+		params->u.lzw.colors = pdf_dict_get_int_default(ctx, p, PDF_NAME(Colors), 1);
+		params->u.lzw.bpc = pdf_dict_get_int_default(ctx, p, PDF_NAME(BitsPerComponent), 8);
+		params->u.lzw.early_change = pdf_dict_get_int_default(ctx, p, PDF_NAME(EarlyChange), 1);
 	}
 	else if (pdf_name_eq(ctx, f, PDF_NAME(JBIG2Decode)))
 	{
@@ -197,7 +182,7 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
  * Create a filter given a name and param dictionary.
  */
 static fz_stream *
-build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params)
+build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_compression_params local_params;
 
@@ -205,6 +190,18 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 	if (params == NULL)
 		params = &local_params;
 
+	if (!might_be_image &&
+		(pdf_name_eq(ctx, f, PDF_NAME(CCITTFaxDecode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(CCF)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(DCTDecode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(DCT)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(JBIG2Decode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(JPXDecode))))
+	{
+		fz_warn(ctx, "Can't open image only stream for non-image purposes");
+		return fz_open_memory(ctx, (unsigned char *)"", 0);
+	}
+
 	build_compression_params(ctx, f, p, params);
 
 	/* If we were using params we were passed in, and we successfully
@@ -257,11 +254,11 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 
 /* Build filter, and assume ownership of chain */
 static fz_stream *
-build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params)
+build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_stream *head;
 	fz_try(ctx)
-		head = build_filter(ctx, tail, doc, f, p, num, gen, params);
+		head = build_filter(ctx, tail, doc, f, p, num, gen, params, might_be_image);
 	fz_always(ctx)
 		fz_drop_stream(ctx, tail);
 	fz_catch(ctx)
@@ -275,7 +272,7 @@ build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *
  * Assume ownership of chain.
  */
 static fz_stream *
-build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params)
+build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_var(chain);
 	fz_try(ctx)
@@ -285,7 +282,7 @@ build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pd
 		{
 			pdf_obj *f = pdf_array_get(ctx, fs, i);
 			pdf_obj *p = pdf_array_get(ctx, ps, i);
-			chain = build_filter_drop(ctx, chain, doc, f, p, num, gen, (i == n-1 ? params : NULL));
+			chain = build_filter_drop(ctx, chain, doc, f, p, num, gen, (i == n-1 ? params : NULL), might_be_image);
 		}
 	}
 	fz_catch(ctx)
@@ -294,9 +291,9 @@ build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pd
 }
 
 static fz_stream *
-build_filter_chain(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params)
+build_filter_chain(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params, int might_be_image)
 {
-	return build_filter_chain_drop(ctx, fz_keep_stream(ctx, chain), doc, fs, ps, num, gen, params);
+	return build_filter_chain_drop(ctx, fz_keep_stream(ctx, chain), doc, fs, ps, num, gen, params, might_be_image);
 }
 
 /*
@@ -313,7 +310,7 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 	pdf_xref_entry *x = NULL;
 	fz_stream *null_stm, *crypt_stm;
 	int hascrypt;
-	int len;
+	int64_t len;
 
 	if (num > 0 && num < pdf_xref_len(ctx, doc))
 	{
@@ -335,10 +332,10 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 	}
 
 	hascrypt = pdf_stream_has_crypt(ctx, stmobj);
-	len = pdf_dict_get_int(ctx, stmobj, PDF_NAME(Length));
+	len = pdf_dict_get_int64(ctx, stmobj, PDF_NAME(Length));
 	if (len < 0)
 		len = 0;
-	null_stm = fz_open_endstream_filter(ctx, file_stm, len, offset);
+	null_stm = fz_open_endstream_filter(ctx, file_stm, (uint64_t)len, offset);
 	if (doc->crypt && !hascrypt)
 	{
 		fz_try(ctx)
@@ -357,7 +354,7 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
  * to stream length and decrypting.
  */
 static fz_stream *
-pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj *stmobj, int num, int64_t offset, fz_compression_params *imparams)
+pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj *stmobj, int num, int64_t offset, fz_compression_params *imparams, int might_be_image)
 {
 	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME(Filter), PDF_NAME(F));
 	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
@@ -368,9 +365,9 @@ pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj
 	fz_try(ctx)
 	{
 		if (pdf_is_name(ctx, filters))
-			fstm = build_filter(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams);
+			fstm = build_filter(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams, might_be_image);
 		else if (pdf_array_len(ctx, filters) > 0)
-			fstm = build_filter_chain(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams);
+			fstm = build_filter_chain(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams, might_be_image);
 		else
 		{
 			if (imparams)
@@ -393,9 +390,9 @@ pdf_open_inline_stream(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj, int
 	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
 
 	if (pdf_is_name(ctx, filters))
-		return build_filter(ctx, file_stm, doc, filters, params, 0, 0, imparams);
+		return build_filter(ctx, file_stm, doc, filters, params, 0, 0, imparams, 1);
 	else if (pdf_array_len(ctx, filters) > 0)
-		return build_filter_chain(ctx, file_stm, doc, filters, params, 0, 0, imparams);
+		return build_filter_chain(ctx, file_stm, doc, filters, params, 0, 0, imparams, 1);
 
 	if (imparams)
 		imparams->type = FZ_IMAGE_RAW;
@@ -415,7 +412,7 @@ pdf_load_compressed_inline_image(fz_context *ctx, pdf_document *doc, pdf_obj *di
 	fz_var(decomp);
 	fz_var(pixmap);
 
-	bc = fz_malloc_struct(ctx, fz_compressed_buffer);
+	bc = fz_new_compressed_buffer(ctx);
 	fz_try(ctx)
 	{
 		bc->buffer = fz_new_buffer(ctx, 1024);
@@ -447,35 +444,35 @@ pdf_open_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 
 	x = pdf_cache_object(ctx, doc, num);
 	if (x->stm_ofs == 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 
 	return pdf_open_raw_filter(ctx, doc->file, doc, x->obj, num, &orig_num, &orig_gen, x->stm_ofs);
 }
 
 static fz_stream *
-pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params)
+pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int might_be_image)
 {
 	pdf_xref_entry *x;
 
 	x = pdf_cache_object(ctx, doc, num);
 	if (x->stm_ofs == 0 && x->stm_buf == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 
-	return pdf_open_filter(ctx, doc, doc->file, x->obj, num, x->stm_ofs, params);
+	return pdf_open_filter(ctx, doc, doc->file, x->obj, num, x->stm_ofs, params, might_be_image);
 }
 
 fz_stream *
 pdf_open_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
-	return pdf_open_image_stream(ctx, doc, num, NULL);
+	return pdf_open_image_stream(ctx, doc, num, NULL, 1);
 }
 
 fz_stream *
 pdf_open_stream_with_offset(fz_context *ctx, pdf_document *doc, int num, pdf_obj *dict, int64_t stm_ofs)
 {
 	if (stm_ofs == 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
-	return pdf_open_filter(ctx, doc, doc->file, dict, num, stm_ofs, NULL);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
+	return pdf_open_filter(ctx, doc, doc->file, dict, num, stm_ofs, NULL, 1);
 }
 
 fz_buffer *
@@ -483,7 +480,7 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
 	fz_stream *stm;
 	pdf_obj *dict;
-	int len;
+	int64_t len;
 	fz_buffer *buf = NULL;
 	pdf_xref_entry *x;
 
@@ -497,7 +494,7 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 	dict = pdf_load_object(ctx, doc, num);
 
 	fz_try(ctx)
-		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+		len = pdf_dict_get_int64(ctx, dict, PDF_NAME(Length));
 	fz_always(ctx)
 		pdf_drop_obj(ctx, dict);
 	fz_catch(ctx)
@@ -505,8 +502,11 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 
 	stm = pdf_open_raw_stream_number(ctx, doc, num);
 
+	if (len < 0)
+		len = 1024;
+
 	fz_try(ctx)
-		buf = fz_read_all(ctx, stm, len);
+		buf = fz_read_all(ctx, stm, (size_t)len);
 	fz_always(ctx)
 		fz_drop_stream(ctx, stm);
 	fz_catch(ctx)
@@ -515,22 +515,30 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 	return buf;
 }
 
-static int
-pdf_guess_filter_length(int len, const char *filter)
+static size_t
+pdf_guess_filter_length(size_t len, const char *filter)
 {
-	if (len < 0)
-		len = 0;
+	size_t nlen = len;
+
+	/* First ones get smaller, no overflow check required. */
 	if (!strcmp(filter, "ASCIIHexDecode"))
 		return len / 2;
-	if (!strcmp(filter, "ASCII85Decode"))
+	else if (!strcmp(filter, "ASCII85Decode"))
 		return len * 4 / 5;
+
 	if (!strcmp(filter, "FlateDecode"))
-		return len * 3;
-	if (!strcmp(filter, "RunLengthDecode"))
-		return len * 3;
-	if (!strcmp(filter, "LZWDecode"))
-		return len * 2;
-	return len;
+		nlen = len * 3;
+	else if (!strcmp(filter, "RunLengthDecode"))
+		nlen = len * 3;
+	else if (!strcmp(filter, "LZWDecode"))
+		nlen = len * 2;
+
+	/* Live with a bad estimate - we'll malloc up as we go, but
+	 * it's probably destined to fail anyway. */
+	if (nlen < len)
+		return len;
+
+	return nlen;
 }
 
 /* Check if an entry has a cached stream and return whether it is directly
@@ -590,7 +598,8 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 {
 	fz_stream *stm = NULL;
 	pdf_obj *dict, *obj;
-	int i, len, n;
+	int i, n;
+	size_t len;
 	fz_buffer *buf;
 
 	fz_var(buf);
@@ -607,12 +616,19 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	dict = pdf_load_object(ctx, doc, num);
 	fz_try(ctx)
 	{
-		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+		int64_t ilen = pdf_dict_get_int64(ctx, dict, PDF_NAME(Length));
+		if (ilen < 0)
+			ilen = 0;
+		len = (size_t)ilen;
+		/* In 32 bit builds, we might find a length being too
+		 * large for a size_t. */
+		if ((int64_t)len != ilen)
+			fz_throw(ctx, FZ_ERROR_LIMIT, "Stream too large");
 		obj = pdf_dict_get(ctx, dict, PDF_NAME(Filter));
 		len = pdf_guess_filter_length(len, pdf_to_name(ctx, obj));
 		n = pdf_array_len(ctx, obj);
 		for (i = 0; i < n; i++)
-			len = pdf_guess_filter_length(len, pdf_to_name(ctx, pdf_array_get(ctx, obj, i)));
+			len = pdf_guess_filter_length(len, pdf_array_get_name(ctx, obj, i));
 	}
 	fz_always(ctx)
 	{
@@ -623,7 +639,7 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 		fz_rethrow(ctx);
 	}
 
-	stm = pdf_open_image_stream(ctx, doc, num, params);
+	stm = pdf_open_image_stream(ctx, doc, num, params, 1);
 
 	fz_try(ctx)
 	{
@@ -650,7 +666,7 @@ pdf_load_stream_number(fz_context *ctx, pdf_document *doc, int num)
 fz_compressed_buffer *
 pdf_load_compressed_stream(fz_context *ctx, pdf_document *doc, int num, size_t worst_case)
 {
-	fz_compressed_buffer *bc = fz_malloc_struct(ctx, fz_compressed_buffer);
+	fz_compressed_buffer *bc = fz_new_compressed_buffer(ctx);
 
 	fz_try(ctx)
 	{
@@ -680,11 +696,12 @@ pdf_open_object_array(fz_context *ctx, pdf_document *doc, pdf_obj *list)
 			fz_concat_push_drop(ctx, stm, pdf_open_stream(ctx, obj));
 		fz_catch(ctx)
 		{
-			if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+			if (fz_caught(ctx) == FZ_ERROR_TRYLATER || fz_caught(ctx) == FZ_ERROR_SYSTEM)
 			{
 				fz_drop_stream(ctx, stm);
 				fz_rethrow(ctx);
 			}
+			fz_report_error(ctx);
 			fz_warn(ctx, "cannot load content stream part %d/%d", i + 1, n);
 		}
 	}
@@ -702,7 +719,7 @@ pdf_open_contents_stream(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 
 	num = pdf_to_num(ctx, obj);
 	if (pdf_is_stream(ctx, obj))
-		return pdf_open_image_stream(ctx, doc, num, NULL);
+		return pdf_open_image_stream(ctx, doc, num, NULL, 0);
 
 	fz_warn(ctx, "content stream is not a stream (%d 0 R)", num);
 	return fz_open_memory(ctx, (unsigned char *)"", 0);
@@ -712,26 +729,26 @@ fz_buffer *pdf_load_raw_stream(fz_context *ctx, pdf_obj *ref)
 {
 	if (pdf_is_stream(ctx, ref))
 		return pdf_load_raw_stream_number(ctx, pdf_get_indirect_document(ctx, ref), pdf_to_num(ctx, ref));
-	fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 }
 
 fz_buffer *pdf_load_stream(fz_context *ctx, pdf_obj *ref)
 {
 	if (pdf_is_stream(ctx, ref))
 		return pdf_load_stream_number(ctx, pdf_get_indirect_document(ctx, ref), pdf_to_num(ctx, ref));
-	fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 }
 
 fz_stream *pdf_open_raw_stream(fz_context *ctx, pdf_obj *ref)
 {
 	if (pdf_is_stream(ctx, ref))
 		return pdf_open_raw_stream_number(ctx, pdf_get_indirect_document(ctx, ref), pdf_to_num(ctx, ref));
-	fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 }
 
 fz_stream *pdf_open_stream(fz_context *ctx, pdf_obj *ref)
 {
 	if (pdf_is_stream(ctx, ref))
 		return pdf_open_stream_number(ctx, pdf_get_indirect_document(ctx, ref), pdf_to_num(ctx, ref));
-	fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "object is not a stream");
 }
